"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("path");
const Bluebird = require("bluebird");
const fs_1 = require("fs");
const util_1 = require("./util");
function readStreamAsync(stream) {
    return new Bluebird(resolve => {
        let input = '';
        stream.setEncoding('utf-8');
        stream.on('data', (x) => {
            input += x;
        });
        stream.once('end', () => resolve(util_1.dequote(input)));
        stream.resume && stream.resume();
    });
}
/**
 * The "cli" step detects whether the process is in a tty. If it is then the input is read into memory.
 * Otherwise, it is buffered from stdin. If no input options are passed in the tty, the package.json#main file is used.
 * After all the build steps have run, the output (the executable) is written to a file or piped to stdout.
 *
 * Configuration:
 *   - compiler.options.input - file path to the input bundle.
 *     - fallbacks: stdin, package.json#main
 *   - compiler.options.output - file path to the output executable.
 *     - fallbacks: stdout, nexe_ + epoch + ext
 *
 * @param {*} compiler
 * @param {*} next
 */
function cli(compiler, next) {
    return __awaiter(this, void 0, void 0, function* () {
        const { input, output } = compiler.options;
        const { log } = compiler;
        if (!input && !process.stdin.isTTY) {
            log.step('Using stdin as input');
            compiler.input = yield readStreamAsync(process.stdin);
        }
        else if (input) {
            log.step(`Using input file as the main module: ${input}`);
            compiler.input = yield util_1.readFileAsync(path_1.normalize(input), 'utf-8');
        }
        else if (!compiler.options.empty) {
            const bundle = require.resolve(process.cwd());
            log.step("Using the cwd's main file as the main module");
            compiler.input = yield util_1.readFileAsync(bundle, 'utf-8');
        }
        else {
            log.step('Using empty input as the main module');
            compiler.input = '';
        }
        yield next();
        const shouldPipeOutput = Boolean(!output && !process.stdout.isTTY);
        const outputName = output || `${compiler.options.name}${util_1.isWindows ? '.exe' : ''}`;
        compiler.output = shouldPipeOutput ? null : outputName;
        const deliverable = yield compiler.compileAsync();
        return new Bluebird((resolve, reject) => {
            deliverable.once('error', reject);
            if (shouldPipeOutput) {
                log.step('Writing binary to stdout');
                deliverable.pipe(process.stdout).once('error', reject);
                resolve();
            }
            else if (compiler.output) {
                const step = log.step('Writing result to file');
                deliverable
                    .pipe(fs_1.createWriteStream(path_1.normalize(compiler.output)))
                    .on('error', reject)
                    .once('close', (e) => {
                    if (e) {
                        reject(e);
                    }
                    else if (compiler.output) {
                        fs_1.chmodSync(compiler.output, '755');
                        step.log(`Executable written to: ${compiler.output}`);
                        resolve(compiler.quit());
                    }
                });
            }
        });
    });
}
exports.default = cli;
