"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = require("path");
const Bluebird = require("bluebird");
const buffer_1 = require("buffer");
const crypto_1 = require("crypto");
const fs_1 = require("fs");
const stream_1 = require("stream");
const child_process_1 = require("child_process");
const nigel_1 = require("nigel");
const logger_1 = require("./logger");
const util_1 = require("./util");
const isBsd = Boolean(~process.platform.indexOf('bsd'));
const make = util_1.isWindows ? 'vcbuild.bat' : isBsd ? 'gmake' : 'make';
const configure = util_1.isWindows ? 'configure' : './configure';
const marker = buffer_1.Buffer.from('<nexe~sentinel>').toString('hex');
const tail = `\n//${marker}`;
const needle = buffer_1.Buffer.from(marker);
const fixedIntegerLength = 10;
const padLeft = (x, l = fixedIntegerLength, c = '0') => (c.repeat(l) + x).slice(-l);
const inflate = (value, size) => {
    if (!size || value.length >= size) {
        return value;
    }
    return value + ' '.repeat(size - value.length);
};
class NexeCompiler {
    constructor(options) {
        this.options = options;
        this.start = Date.now();
        this.env = Object.assign({}, process.env);
        this.log = new logger_1.Logger(this.options.loglevel);
        this.src = path_1.join(this.options.temp, this.options.version);
        this.files = [];
        this.resources = {
            index: {},
            bundle: ''
        };
        this.nodeSrcBinPath = util_1.isWindows
            ? path_1.join(this.src, 'Release', 'node.exe')
            : path_1.join(this.src, 'out', 'Release', 'node');
        const { python } = (this.options = options);
        if (python) {
            if (util_1.isWindows) {
                this.env.PATH = '"' + util_1.dequote(path_1.normalize(python)) + '";' + this.env.PATH;
            }
            else {
                this.env.PYTHON = python;
            }
        }
        this.readFileAsync = (file) => __awaiter(this, void 0, void 0, function* () {
            let cachedFile = this.files.find(x => path_1.normalize(x.filename) === path_1.normalize(file));
            if (!cachedFile) {
                cachedFile = {
                    filename: file,
                    contents: yield util_1.readFileAsync(path_1.join(this.src, file), 'utf-8').catch({ code: 'ENOENT' }, () => '')
                };
                this.files.push(cachedFile);
            }
            return cachedFile;
        });
        this.writeFileAsync = (file, contents) => util_1.writeFileAsync(path_1.join(this.src, file), contents);
        this.replaceInFileAsync = (file, replace, value) => __awaiter(this, void 0, void 0, function* () {
            const entry = yield this.readFileAsync(file);
            entry.contents = entry.contents.replace(replace, value);
        });
        this.setFileContentsAsync = (file, contents) => __awaiter(this, void 0, void 0, function* () {
            const entry = yield this.readFileAsync(file);
            entry.contents = contents;
        });
    }
    quit(code = 0) {
        const time = Date.now() - this.start;
        this.log.write(`Finsihed in ${time / 1000}s`);
        return this.log.flush().then(x => process.exit(code));
    }
    _findPaddingSize(size, override = this.options.padding) {
        if (override === 0) {
            return 0;
        }
        size = override > size ? override : size;
        const padding = [3, 6, 9, 16, 25, 40].map(x => x * 1e6).find(p => size <= p);
        if (!padding) {
            throw new Error(`No prebuilt target large enough (${(size / 1024).toFixed(2)}Mb).\nUse the --build flag and build for the current platform`);
        }
        return padding;
    }
    _getNodeExecutableLocation(target) {
        if (target) {
            return path_1.join(this.options.temp, target);
        }
        return this.nodeSrcBinPath;
    }
    _runBuildCommandAsync(command, args) {
        return new Bluebird((resolve, reject) => {
            child_process_1.spawn(command, args, {
                cwd: this.src,
                env: this.env,
                stdio: 'ignore'
            })
                .once('error', reject)
                .once('close', resolve);
        });
    }
    _configureAsync() {
        return this._runBuildCommandAsync(this.env.PYTHON || 'python', [
            configure,
            ...this.options.configure
        ]);
    }
    _buildAsync() {
        return __awaiter(this, void 0, void 0, function* () {
            this.compileStep.log(`Configuring node build: ${this.options.configure}`);
            yield this._configureAsync();
            const buildOptions = util_1.isWindows ? this.options.vcBuild : this.options.make;
            this.compileStep.log(`Compiling Node with arguments: ${buildOptions}`);
            yield this._runBuildCommandAsync(make, buildOptions);
            return fs_1.createReadStream(this._getNodeExecutableLocation());
        });
    }
    _fetchPrebuiltBinaryAsync() {
        return this._buildAsync();
    }
    _getPayload(header) {
        return this._serializeHeader(header) + this.input + '/**' + this.resources.bundle + `**/`;
    }
    _generateHeader() {
        const zeros = padLeft(0);
        const version = ['configure', 'vcBuild', 'make'].reduce((a, c) => {
            return (a += this.options[c].slice().sort().join());
        }, '') + this.options.enableNodeCli;
        const header = {
            version: padLeft(0, 32),
            resources: this.resources.index,
            contentSize: zeros,
            paddingSize: zeros,
            resourceOffset: zeros,
            binaryOffset: zeros
        };
        const serializedHeader = this._serializeHeader(header);
        header.contentSize = padLeft(buffer_1.Buffer.byteLength(this._getPayload(header)));
        header.paddingSize = padLeft(this._findPaddingSize(+header.contentSize));
        header.resourceOffset = padLeft(buffer_1.Buffer.byteLength(serializedHeader + this.input + '/**'));
        header.version = crypto_1.createHash('md5').update(version + header.paddingSize).digest('hex');
        return header;
    }
    _getExistingBinaryHeaderAsync(target) {
        return __awaiter(this, void 0, void 0, function* () {
            const filename = this._getNodeExecutableLocation(target);
            const existingBinary = yield util_1.pathExistsAsync(filename);
            if (existingBinary) {
                return this._extractHeaderAsync(filename);
            }
            return null;
        });
    }
    _extractHeaderAsync(path) {
        const binary = fs_1.createReadStream(path);
        const haystack = new nigel_1.Stream(needle);
        let needles = 0;
        let stackCache = [];
        return new Promise((resolve, reject) => {
            binary
                .on('error', reject)
                .pipe(haystack)
                .on('error', reject)
                .on('close', () => reject(new Error(`Binary: ${path} is not compatible with nexe`)))
                .on('haystack', (x) => needles && stackCache.push(x))
                .on('needle', () => {
                if (++needles === 2) {
                    resolve(JSON.parse(buffer_1.Buffer.concat(stackCache).toString()));
                    binary.close();
                    haystack.end();
                }
            });
        });
    }
    _serializeHeader(header) {
        return `/**${marker}${JSON.stringify(header)}${marker}**/process.__nexe=${JSON.stringify(header)};`;
    }
    setMainModule(compiler, next) {
        return __awaiter(this, void 0, void 0, function* () {
            yield next();
            const header = compiler._generateHeader();
            const contents = inflate(this._getPayload(header), +header.paddingSize) + tail;
            return compiler.setFileContentsAsync(`lib/${compiler.options.name}.js`, contents);
        });
    }
    compileAsync() {
        return __awaiter(this, void 0, void 0, function* () {
            const step = (this.compileStep = this.log.step('Compiling result'));
            let target = this.options.targets.slice().shift();
            let prebuiltBinary = null;
            const header = this._generateHeader();
            target = target && `${target}-${header.version.slice(0, 6)}`;
            step.log(`Scanning existing binary...`);
            const existingHeader = yield this._getExistingBinaryHeaderAsync(target);
            if (existingHeader && existingHeader.version === header.version) {
                const location = this._getNodeExecutableLocation(target);
                step.log(`Source already built: ${location}`);
                prebuiltBinary = fs_1.createReadStream(location);
            }
            if (target) {
                throw new Error('\nNot Implemented, use --build during beta\n');
                // prebuiltBinary = await this._fetchPrebuiltBinaryAsync(target)
            }
            if (!prebuiltBinary) {
                prebuiltBinary = yield this._buildAsync();
                step.log('Node binary compiled');
            }
            return this._assembleDeliverable(header, prebuiltBinary);
        });
    }
    _assembleDeliverable(header, binary) {
        const haystack = new nigel_1.Stream(buffer_1.Buffer.concat([buffer_1.Buffer.from('/**'), needle]));
        const artifact = new stream_1.Readable({ read() { } });
        let needles = 0;
        let currentStackSize = 0;
        binary.pipe(haystack);
        haystack
            .on('close', () => artifact.push(null))
            .on('needle', () => ++needles && haystack.needle(needle))
            .on('haystack', (x) => {
            if (!needles) {
                currentStackSize += x.length;
                artifact.push(x);
            }
            if (needles === 1 && !+header.binaryOffset) {
                header.binaryOffset = padLeft(currentStackSize);
                const content = buffer_1.Buffer.from(inflate(this._getPayload(header), +header.paddingSize) + tail);
                artifact.push(content);
            }
            if (needles > 2) {
                artifact.push(x);
            }
        });
        return artifact;
    }
}
exports.NexeCompiler = NexeCompiler;
